#' Helper function to check for GRangesList
#' @param class the class you want to check if is GRL,
#' either a character from class or the object itself.
#' @return a boolean
#' @family utils
#'
is.grl <- function(class) {
  if (!is.character(class)) {
    class <- class(class)
  }
  return((class == "GRangesList" || class == "CompressedGRangesList"))
}


#' Helper function to check for GRangesList or GRanges class
#' @param class the class you want to check if is GRL or GR,
#'  either a character from class or the object itself.
#' @return a boolean
#' @family utils
#'
is.gr_or_grl <- function(class) {
  if (!is.character(class)) {
    class <- class(class)
  }
  return(is.grl(class) || class == "GRanges")
}


#' Helper Function to check valid GRangesList input
#' @param class as character vector the given class of
#'  supposed GRangesList object
#' @param type a character vector, is it gtf, cds, 5', 3', for messages.
#' @param checkNULL should NULL classes be checked and return indeces of these?
#' @return either NULL or indices (checkNULL == TRUE)
#' @family utils
#'
validGRL <- function(class, type = "grl", checkNULL = FALSE) {
  if(length(class) != length(type)) stop("not equal length of classes",
                                         " and types, see validGRL")
  if (checkNULL) {
    indeces <- "NULL" == class
    class <- class[!indeces]
    if (length(class) == 0) return(rep(TRUE, length(type)))
    type <- type[!indeces]
  }
  for (classI in seq_along(class)) {
    if (!is.grl(class[classI])) {
      messageI <- paste(type[classI], "must be given and be type GRangesList")
      stop(messageI)
    }
  }
  if (checkNULL) {
    return(indeces)
  }
}

#' Converts different type of files to Granges
#'
#' column 5 will be set to score
#' Only Accepts bed files for now, standard format from Fantom5
#' @param x An data.frame from imported bed-file,
#'  to convert to GRanges
#' @param bed6 If bed6, no meta column is added
#' @return a GRanges object from bed
#' @family utils
#'
bedToGR <- function(x, bed6 = TRUE){

  if (!bed6) {
    gr <- GRanges(x[, 1], IRanges(x[, 2] + 1, x[, 3]))
    return(gr)
  }
  starts <- x[, 2] + 1
  ends <- x[, 3]
  gr <- GRanges(x[, 1], IRanges(starts, ends),
                strand = x[, 6])
  mcols(gr) <- S4Vectors::DataFrame(mcols(gr), score = x[, 5])
  if (ncol(x) > 6) mcols(gr) <- x[, 7:ncol(x)]
  return(gr)
}


#' Load bed file as GRanges.
#'
#' Wraps around rtracklayer::import.bed and tries to speed up loading with the
#' use of data.table. Supports gzip, gz, bgz and bed formats.
#' @param filePath The location of the bed file
#' @importFrom data.table fread setDF
#' @importFrom tools file_ext
#' @importFrom rtracklayer import.bed
#' @return a GRanges object
#' @export
#' @family utils
#' @examples
#' # path to example CageSeq data from hg19 heart sample
#' cageData <- system.file("extdata", "cage-seq-heart.bed.bgz",
#'                         package = "ORFik")
#' fread.bed(cageData)
#'
fread.bed <- function(filePath) {

  if (.Platform$OS.type == "unix") {
    if (file.exists(filePath)) {
      if (any(file_ext(filePath) == c("gzip", "gz", "bgz"))) {
        bed <- bedToGR(setDF(
          fread(cmd = paste("gunzip -c", filePath), sep = "\t")))
      } else if (file_ext(filePath) == "bed"){
        bed <- bedToGR(setDF(fread(filePath, sep = "\t")))
      } else {
        bed <- import.bed(con =  filePath)
      }
    } else {stop("Filepath specified does not name existing file.")}
  } else {
    ## NB: Windows user will have slower loading
    bed <- import.bed(con =  filePath)
  }
  return(bed)
}

#' Convert a GRanges Object to 1 width reads
#'
#' There are 3 ways of doing this
#' 1. Take 5' ends, reduce away rest (5prime)
#' 2. Tile and include all (tileAll)
#' 3. Take middle point per GRanges (middle)
#' @param gr GRanges Object to reduce
#' @param method the method to reduce, see info. (5prime defualt)
#' @param addScoreColumn logical (FALSE), if TRUE, add a score column that
#'  sums up the hits per position.
#' @return  Converted GRanges object
#' @family utils
#'
convertToOneBasedRanges <- function(gr, method = "5prime",
                                    addScoreColumn = FALSE){
  if (method == "5prime") {
    gr <- resize(gr, width = 1)
  } else if(method == "tileAll") {
    gr <- unlist(tile(gr, width = 1), use.names = FALSE)

  } else if(method == "middle") {
    ranges(gr) <- IRanges(start(gr) + ceiling((end(gr)-start(gr))/2),
                          width = 1)
  } else stop("method not defined")

  if (addScoreColumn) {
    pos <- strandBool(gr)
    posGr <- gr[pos]
    dt <- as.data.table(posGr)[, .N, .(seqnames, start)]
    posGr <- GRanges(seqnames = dt$seqnames, IRanges(dt$start, width = 1),
                  strand = dt$strand)
    score <- dt$N
    negGr <- gr[!pos]
    dt <- as.data.table(posGr)[, .N, .(seqnames, end)]
    negGr <- GRanges(seqnames = dt$seqnames, IRanges(dt$end, width = 1),
                     strand = dt$strand)
    score <- as.integer(c(score, dt$N))

    gr <- c(posGr, negGr)
    gr$score <- NULL
    mcols(gr) <- S4Vectors::DataFrame(mcols(gr), score = score)
  }

  return(gr)
}

#' Convenience wrapper for Rsamtools FaFile
#' @param faFile a character path or FaFile
#' @importFrom Rsamtools FaFile
#' @return a FaFile or BSgenome
#' @family utils
#'
findFa <- function(faFile) {
  if (is.character(faFile)) {
    if (dir.exists(faFile)) {
      return(FaFile(faFile))
    } else {
      stop("faFile does not name a valid fasta file")
    }

  } else if (is(faFile, "FaFile") || is(faFile, "BSgenome")) {
    return(faFile)
  }
  stop("faFile must be FaFile, BSgenome or valid filePath")
}
